use std::io::{Read, Result};

use serde::Deserialize;
use skyward_mavlink::lyra::{
    MavMessage, PAYLOAD_FLIGHT_TM_DATA, PAYLOAD_STATS_TM_DATA, ROCKET_FLIGHT_TM_DATA,
    ROCKET_STATS_TM_DATA,
};

use crate::packet::TimedMessage;

#[derive(Debug, Deserialize)]
struct Lrtm {
    timestamp: u64,
    latitude: f64,
    longitude: f64,
    altitude: f64,
}

#[derive(Debug, Deserialize)]
struct HrtmMain {
    timestamp: u64,
    n: f64,
    e: f64,
    d: f64,
    vn: f64,
    ve: f64,
    vd: f64,
    #[serde(rename = "aglAltitude")]
    altitude_agl: f64,
    #[serde(rename = "verticalSpeed")]
    vertical_speed: f64,
}

#[derive(Debug, Deserialize)]
struct HrtmPayload {
    timestamp: u64,
    n: f64,
    e: f64,
    d: f64,
    vn: f64,
    ve: f64,
    vd: f64,
}

pub enum Origin {
    Main,
    Payload,
}

/// Read low rate telemetry data from a CSV file
pub fn read_low_rate_telemetry(reader: impl Read, origin: Origin) -> Result<Vec<TimedMessage>> {
    let mut packets = Vec::new();
    let mut rdr = csv::Reader::from_reader(reader);
    for result in rdr.deserialize() {
        let r: Lrtm = result?;
        match origin {
            Origin::Main => {
                let msg_data = ROCKET_STATS_TM_DATA {
                    ref_lat: r.latitude as f32,
                    ref_lon: r.longitude as f32,
                    ref_alt: r.altitude as f32,
                    ..Default::default()
                };
                let mav_msg = MavMessage::ROCKET_STATS_TM(msg_data);
                packets.push(TimedMessage::new(r.timestamp, mav_msg));
            }
            Origin::Payload => {
                let msg_data = PAYLOAD_STATS_TM_DATA {
                    ref_lat: r.latitude as f32,
                    ref_lon: r.longitude as f32,
                    ref_alt: r.altitude as f32,
                    ..Default::default()
                };
                let mav_msg = MavMessage::PAYLOAD_STATS_TM(msg_data);
                packets.push(TimedMessage::new(r.timestamp, mav_msg));
            }
        }
    }
    Ok(packets)
}

/// Read high rate telemtry data from a CSV file
pub fn read_high_rate_telemetry(
    reader_main: impl Read,
    reader_payload: impl Read,
) -> Result<Vec<TimedMessage>> {
    let mut packets = Vec::new();
    let mut rdr_main = csv::Reader::from_reader(reader_main);
    let mut rdr_payload = csv::Reader::from_reader(reader_payload);
    for (main_res, payload_res) in rdr_main.deserialize().zip(rdr_payload.deserialize()) {
        let rm: HrtmMain = main_res?;
        let msg_data = ROCKET_FLIGHT_TM_DATA {
            timestamp: rm.timestamp,
            nas_n: rm.n as f32,
            nas_e: rm.e as f32,
            nas_d: rm.d as f32,
            nas_vn: rm.vn as f32,
            nas_ve: rm.ve as f32,
            nas_vd: rm.vd as f32,
            altitude_agl: rm.altitude_agl as f32,
            ada_vert_speed: rm.vertical_speed as f32,
            ..Default::default()
        };
        let mav_msg = MavMessage::ROCKET_FLIGHT_TM(msg_data);
        packets.push(TimedMessage::new(rm.timestamp, mav_msg));

        let rp: HrtmPayload = payload_res?;
        let msg_data = PAYLOAD_FLIGHT_TM_DATA {
            timestamp: rp.timestamp,
            nas_n: rp.n as f32,
            nas_e: rp.e as f32,
            nas_d: rp.d as f32,
            nas_vn: rp.vn as f32,
            nas_ve: rp.ve as f32,
            nas_vd: rp.vd as f32,
            ..Default::default()
        };
        let mav_msg = MavMessage::PAYLOAD_FLIGHT_TM(msg_data);
        packets.push(TimedMessage::new(rm.timestamp, mav_msg));
    }
    Ok(packets)
}
