import os
from pathlib import Path

import click
import polars as pl

NAS_CALIBRATE_STATE = 1
MAIN_FLYING_FMM_STATE = 10
PAYLOAD_FLYING_FMM_STATE = 9
SECONDS_BEFORE_FLYING = 15


@click.command()
@click.argument("path", type=click.Path(exists=True))
@click.option("--output", type=click.Path(), default=".")
@click.option("--origin-main", "origin", flag_value="main", default=True)
@click.option("--origin-payload", "origin", flag_value="payload")
def main(path: Path, output: Path, origin: str):
    """
    Convert the logs found in the given path to a format compliant with ARPIST.
    """
    # now walk in the directory pointed by path the files: main_Boardcore_EventData.csv, main_Boardcore_NASState.csv, main_Boardcore_ReferenceValues.csv and save their path to variables
    nas_controller_status_path = None
    nas_state_path = None
    reference_values_path = None
    fmm_status_path = None
    gps_path = None

    for p, _, fn in os.walk(path):
        for f in fn:
            if origin == "main":
                if f == "main_Main_NASControllerStatus.csv":
                    nas_controller_status_path = os.path.join(p, f)
                elif f == "main_Main_FlightModeManagerStatus.csv":
                    fmm_status_path = os.path.join(p, f)
                elif f == "main_Boardcore_NASState.csv":
                    nas_state_path = os.path.join(p, f)
            elif origin == "payload":
                if f == "payload_Payload_NASControllerStatus.csv":
                    nas_controller_status_path = os.path.join(p, f)
                elif f == "payload_Payload_FlightModeManagerStatus.csv":
                    fmm_status_path = os.path.join(p, f)
                elif f == "payload_Boardcore_NASState.csv":
                    nas_state_path = os.path.join(p, f)
                elif f == "payload_Boardcore_UBXGPSData.csv":
                    gps_path = os.path.join(p, f)
            if f == "main_Boardcore_ReferenceValues.csv":
                reference_values_path = os.path.join(p, f)

    if nas_controller_status_path is None:
        raise FileNotFoundError("NAS controller status file not found")
    if nas_state_path is None:
        raise FileNotFoundError("NAS state file not found")
    if reference_values_path is None:
        raise FileNotFoundError("Reference values file not found")
    if fmm_status_path is None:
        raise FileNotFoundError("FMM status file not found")
    if origin == "payload" and gps_path is None:
        raise FileNotFoundError("GPS data file not found")

    nas_controller_status = pl.read_csv(nas_controller_status_path)
    nas_state = pl.read_csv(nas_state_path)
    reference_values = pl.read_csv(reference_values_path)
    fmm_status = pl.read_csv(fmm_status_path)

    # sort by timestamp and extract the timestamp associated to the calibrate event and topic
    nas_controller_status = nas_controller_status.sort("timestamp")
    calibrate_tms = nas_controller_status.filter(
        pl.col("state") == NAS_CALIBRATE_STATE
    ).select("timestamp")

    # add the calibrate timestamp to the reference_values as a new column
    reference_values = reference_values.with_columns(calibrate_tms.select("timestamp"))

    # select cols
    reference_values = reference_values.select(
        pl.from_epoch(pl.col("timestamp"), time_unit="us"),
        pl.col("refLatitude").alias("latitude"),
        pl.col("refLongitude").alias("longitude"),
        pl.col("refAltitude").alias("altitude"),
    )
    nas_state = nas_state.select(
        pl.from_epoch(pl.col("timestamp"), time_unit="us"),
        "n",
        "e",
        "d",
        "vn",
        "ve",
        "vd",
    )
    fmm_status = fmm_status.select(
        pl.from_epoch(pl.col("timestamp"), time_unit="us"), "state"
    )

    # find stop timestamp based on last state of the fmm plus 10 seconds
    stop_ts = fmm_status.select(pl.col("timestamp") + pl.duration(seconds=2))[-1, 0]

    # filter the reference values and nas state dataframes
    reference_values = reference_values.filter(pl.col("timestamp") <= stop_ts)
    nas_state = nas_state.filter(pl.col("timestamp") <= stop_ts)

    # apply correction to the reference values if payload origin
    if origin == "payload":
        gps_data = pl.read_csv(gps_path, infer_schema_length=10000)
        gps_data = gps_data.select(
            pl.from_epoch(pl.col("gpsTimestamp"), time_unit="us").alias("timestamp"),
            "latitude",
            "longitude",
        )
        new_reference_values = []
        for row in reference_values.rows():
            timestamp = row[0]
            (lat, lon) = (
                gps_data.filter(pl.col("timestamp") < timestamp)
                .tail(1)
                .to_numpy()[0, 1:]
            )
            new_reference_values.append(
                [timestamp, lat, lon, row[3] + 1.5]  # altitude is corrected
            )
        reference_values = pl.DataFrame(
            new_reference_values, schema=reference_values.schema, orient="row"
        )

    # find max timestamp
    max_ts = max(
        reference_values.select("timestamp").max().item(0, 0),
        (nas_state.select("timestamp").max().item(0, 0)),
    )

    # upsample and downsample the dataframes
    last_row = reference_values.tail(1)
    last_row[0, "timestamp"] = max_ts
    reference_values = pl.concat([reference_values, last_row], how="vertical")
    reference_values = (
        reference_values.group_by_dynamic(pl.col("timestamp"), every="500ms")
        .agg(pl.all().last())
        .upsample(time_column="timestamp", every="500ms")
        .fill_null(strategy="forward")
    )
    nas_state = (
        nas_state.group_by_dynamic(pl.col("timestamp"), every="250ms")
        .agg(pl.all().last())
        .upsample(time_column="timestamp", every="250ms")
        .fill_null(strategy="forward")
    )

    # filter from 15 seconds before flying
    flying_fmm_state = (
        MAIN_FLYING_FMM_STATE if origin == "main" else PAYLOAD_FLYING_FMM_STATE
    )
    start_ts = fmm_status.filter(pl.col("state") == flying_fmm_state).select(
        "timestamp"
    )[0, 0] - pl.duration(seconds=SECONDS_BEFORE_FLYING)
    reference_values = reference_values.filter(pl.col("timestamp") >= start_ts)
    nas_state = nas_state.filter(pl.col("timestamp") >= start_ts)

    # save the dataframes to csv
    output = Path(output)
    reference_values.select(
        pl.col("timestamp").dt.timestamp(time_unit="us"),
        "latitude",
        "longitude",
        "altitude",
    ).write_csv(output / "low_rate.csv")
    nas_state.select(
        pl.col("timestamp").dt.timestamp(time_unit="us"),
        "n",
        "e",
        "d",
        "vn",
        "ve",
        "vd",
    ).write_csv(output / "high_rate.csv")


if __name__ == "__main__":
    main()
